"""
Probabilistic machine learning module for Python
=================================================
pmlearn is a Python module for practical probabilistic machine
learning built on top of scikit-learn and PymC3.

It aims to provide simple and efficient solutions to learning problems
that are accessible to everybody and reusable in various contexts:
machine-learning as a versatile tool for science and engineering.
See http://pymc-learn.org for complete documentation.
"""

__version__ = '0.0.1.rc2'

__all__ = ['gaussian_process',
           'linear_model',
           'mixture',
           'neural_network',
           'naive_bayes']
